package ru.dbotthepony.mc.otm.worldgen.placement

import com.mojang.serialization.Codec
import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.BlockPos
import net.minecraft.util.RandomSource
import net.minecraft.util.valueproviders.ClampedNormalFloat
import net.minecraft.util.valueproviders.ConstantFloat
import net.minecraft.util.valueproviders.FloatProvider
import net.minecraft.util.valueproviders.IntProvider
import net.minecraft.util.valueproviders.UniformFloat
import net.minecraft.world.level.levelgen.placement.PlacementContext
import net.minecraft.world.level.levelgen.placement.PlacementModifier
import net.minecraft.world.level.levelgen.placement.PlacementModifierType
import ru.dbotthepony.mc.otm.core.addAll
import ru.dbotthepony.mc.otm.core.math.Vector
import ru.dbotthepony.mc.otm.core.math.angleDifference
import ru.dbotthepony.mc.otm.core.math.normalizeAngle
import ru.dbotthepony.mc.otm.core.math.plus
import ru.dbotthepony.mc.otm.core.math.toBlockPos
import ru.dbotthepony.mc.otm.core.nextDouble
import ru.dbotthepony.mc.otm.data.codec.minRange
import ru.dbotthepony.mc.otm.data.world.BooleanProvider
import ru.dbotthepony.mc.otm.registry.data.MPlacementModifiers
import ru.dbotthepony.mc.otm.worldgen.EnhancedPlacementContext
import java.util.stream.Stream
import kotlin.math.PI
import kotlin.math.absoluteValue
import kotlin.math.cos
import kotlin.math.sign
import kotlin.math.sin

/**
 * Creates one block thick "worm" positions which twists and turns
 */
class WormPlacement(
	val length: IntProvider,
	val turnChanceXZ: BooleanProvider,
	val turnChanceXY: BooleanProvider,
	val turnSpeedXZ: FloatProvider,
	val turnSpeedXY: FloatProvider,
	val turnRateXZ: FloatProvider,
	val turnRateXY: FloatProvider,
	val initialAngleXY: FloatProvider = DEFAULT_INITIAL_ANGLE_XY,
	val maxTravelDown: Int = Int.MAX_VALUE,
	val maxTravelUp: Int = Int.MAX_VALUE,
) : PlacementModifier(), EnhancedPlacementModifier {
	private inner class Worm(private val random: RandomSource, private var position: Vector) {
		private var remainingDistance = length.sample(random)
		private var xzRotation = random.nextDouble(-PI, PI)
		private var xyRotation = normalizeAngle(initialAngleXY.sample(random).toDouble())
		private var xzTargetRotation = random.nextDouble(-PI, PI)
		private var xyTargetRotation = normalizeAngle(initialAngleXY.sample(random).toDouble())

		private var xzSin = sin(xzRotation)
		private var xzCos = cos(xzRotation)
		private var xySin = sin(xyRotation)

		private val turnChanceXZ = this@WormPlacement.turnChanceXZ.instance()
		private val turnChanceXY = this@WormPlacement.turnChanceXY.instance()

		val hasFinished: Boolean
			get() = remainingDistance <= 0

		private var prevPos = BlockPos.ZERO

		fun follow(): BlockPos? {
			if (hasFinished) return null
			remainingDistance--

			// wormy turn
			if (turnChanceXZ.sample(random)) {
				// wormy angle
				xzTargetRotation += turnRateXZ.sample(random)
				xzTargetRotation = normalizeAngle(xzTargetRotation)
			}

			// wormy-o!
			if (turnChanceXY.sample(random)) {
				// worm?
				xyTargetRotation += turnRateXY.sample(random)
				xyTargetRotation = normalizeAngle(xyTargetRotation)
			}

			if (xzTargetRotation != xzRotation) {
				val diff = angleDifference(xzTargetRotation, xzRotation)
				val abs = diff.absoluteValue
				val speed = turnSpeedXZ.sample(random).toDouble()

				if (abs <= speed)
					xzRotation = xzTargetRotation
				else
					xzRotation += speed * diff.sign

				xzSin = sin(xzRotation)
				xzCos = cos(xzRotation)
			}

			if (xyTargetRotation != xyRotation) {
				val diff = angleDifference(xyTargetRotation, xyRotation)
				val abs = diff.absoluteValue
				val speed = turnSpeedXZ.sample(random).toDouble()

				if (abs <= speed)
					xyRotation = xyTargetRotation
				else
					xyRotation += speed * diff.sign

				xySin = sin(xyRotation)
			}

			// advance worm
			position += Vector(xzCos, xySin, xzSin)

			if (position.y > maxTravelUp) {
				xyTargetRotation = 0.0
				xySin = 0.0
				position = Vector(position.x, maxTravelUp.toDouble(), position.z)
			} else if (position.y < -maxTravelDown) {
				xyTargetRotation = 0.0
				xySin = 0.0
				position = Vector(position.x, -maxTravelDown.toDouble(), position.z)
			}

			val calc = position.toBlockPos()

			if (calc != prevPos) {
				// if worm made a wurm, add new position to set
				prevPos = calc
				return calc
			}

			return null
		}
	}

	private fun evaluate(random: RandomSource, position: BlockPos, results: ArrayList<BlockPos>) {
		val worms = ArrayList<Worm>()
		worms.add(Worm(random, Vector.ZERO))
		results.add(position)

		while (worms.isNotEmpty()) {
			worms.removeIf {
				val pos = it.follow()
				if (pos != null) results.add(pos + position)
				it.hasFinished
			}
		}
	}

	override fun evaluate(context: EnhancedPlacementContext, positions: List<BlockPos>): List<BlockPos> {
		if (positions.isEmpty())
			return positions
		else {
			val results = ArrayList<BlockPos>()
			positions.forEach { evaluate(context.random, it, results) }
			return results
		}
	}

	override val type: EnhancedPlacementModifier.Type<*>
		get() = Companion

	override fun getPositions(context: PlacementContext, random: RandomSource, center: BlockPos): Stream<BlockPos> {
		val results = ArrayList<BlockPos>()
		evaluate(random, center, results)
		return results.stream()
	}

	override fun type(): PlacementModifierType<*> {
		return MPlacementModifiers.WORM_PLACEMENT
	}

	companion object : EnhancedPlacementModifier.Type<WormPlacement> {
		private fun increment(value: Float): Float {
			var i = 1f

			while (true) {
				val enlarge = Float.MIN_VALUE * i
				val sum = value + enlarge

				if (sum != value) {
					return sum
				} else {
					i *= 2f
				}
			}
		}

		val DEFAULT_INITIAL_ANGLE_XY: UniformFloat = UniformFloat.of(-PI.toFloat() / 16f, increment(PI.toFloat() / 16f))

		fun constantTurnRate(degrees: Float): FloatProvider {
			return constantTurnRateRadians(degrees * DEGREES_TO_RADIANS)
		}

		fun constantTurnRateRadians(rad: Float): FloatProvider {
			return ConstantFloat.of(rad)
		}

		fun uniformTurnRate(degrees: Float): FloatProvider {
			return uniformTurnRateRadians(degrees * DEGREES_TO_RADIANS)
		}

		fun uniformTurnRateRadians(rad: Float): FloatProvider {
			require(rad >= 0f) { "Provided value must be non-negative, $rad given" }
			if (rad == 0f) return ConstantFloat.ZERO
			return UniformFloat.of(-rad, increment(rad))
		}

		fun normalDistributedTurnRate(deviation: Float, mean: Float = 0f): FloatProvider {
			return normalDistributedTurnRateRadians(deviation * DEGREES_TO_RADIANS, mean * DEGREES_TO_RADIANS)
		}

		fun normalDistributedTurnRateRadians(deviation: Float, mean: Float = 0f): FloatProvider {
			require(deviation >= 0f) { "Provided value must be non-negative, $deviation given" }
			if (deviation == 0f) return ConstantFloat.ZERO
			return ClampedNormalFloat.of(mean, deviation, -PI.toFloat() * 2f, PI.toFloat() * 2f)
		}

		private const val DEGREES_TO_RADIANS = PI.toFloat() / 180f

		val CODEC: MapCodec<WormPlacement> by lazy {
			RecordCodecBuilder.mapCodec {
				it.group(
					IntProvider.CODEC.fieldOf("length").forGetter(WormPlacement::length),
					BooleanProvider.CODEC.fieldOf("turn_chance_xz").forGetter(WormPlacement::turnChanceXZ),
					BooleanProvider.CODEC.fieldOf("turn_chance_xy").forGetter(WormPlacement::turnChanceXY),
					FloatProvider.CODEC.fieldOf("turn_speed_xz").forGetter(WormPlacement::turnSpeedXZ),
					FloatProvider.CODEC.fieldOf("turn_speed_xy").forGetter(WormPlacement::turnSpeedXY),
					FloatProvider.CODEC.fieldOf("turn_rate_xz").forGetter(WormPlacement::turnRateXZ),
					FloatProvider.CODEC.fieldOf("turn_rate_xy").forGetter(WormPlacement::turnRateXY),
					FloatProvider.CODEC.optionalFieldOf("initial_angle_xy", DEFAULT_INITIAL_ANGLE_XY).forGetter(WormPlacement::initialAngleXY),
					Codec.INT.minRange(1).optionalFieldOf("max_travel_down", Int.MAX_VALUE).forGetter(WormPlacement::maxTravelDown),
					Codec.INT.minRange(1).optionalFieldOf("max_travel_up", Int.MAX_VALUE).forGetter(WormPlacement::maxTravelUp),
				).apply(it, ::WormPlacement)
			}
		}

		override val codec: MapCodec<WormPlacement>
			get() = CODEC
	}
}
